@@ -21,10 +21,12 @@
 
 #include "config.h"
 
+#include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
+#include <sys/stat.h>
 #include <sys/types.h>
 #include <langinfo.h>
 #include <errno.h>
@@ -39,6 +41,9 @@
 
 #include "terminal-accels.hh"
 #include "terminal-app.hh"
+#include "terminal-client-utils.hh"
+#include "terminal-debug.hh"
+#include "terminal-defines.hh"
 #include "terminal-intl.hh"
 #include "terminal-util.hh"
 #include "terminal-version.hh"
@@ -992,8 +997,8 @@ terminal_util_number_info (const char *str)
 
   errno = 0;
   char* end;
-  gint64 num = g_ascii_strtoull(str, &end, hex ? 16 : 10);
-  if (errno || str == end || num == -1)
+  guint64 num = g_ascii_strtoull(str, &end, hex ? 16 : 10);
+  if (errno || str == end)
     return nullptr;
 
   /* No use in dec-hex conversion for so small numbers */
@@ -1006,18 +1011,18 @@ terminal_util_number_info (const char *str)
   if (thousep[0] != '\0') {
     /* If thousep is nonempty, use printf's magic which can handle
        more complex separating logics, e.g. 2+2+2+3 for some locales */
-    decstr = g_strdup_printf("%'" G_GINT64_FORMAT, num);
+    decstr = g_strdup_printf("%'" G_GUINT64_FORMAT, num);
   } else {
     /* If, however, thousep is empty, override it with a space so that we
        do always group the digits (that's the whole point of this feature;
        the choice of space guarantees not conflicting with the decimal separator) */
-    gs_free char *tmp = g_strdup_printf("%" G_GINT64_FORMAT, num);
+    gs_free char *tmp = g_strdup_printf("%" G_GUINT64_FORMAT, num);
     thousep = " ";
     decstr = add_separators(tmp, thousep, 3);
   }
 
   /* Group the hex digits by 4 using the same nonempty separator */
-  hextmp = g_strdup_printf("%" G_GINT64_MODIFIER "x", (guint64)(num));
+  hextmp = g_strdup_printf("%" G_GINT64_MODIFIER "x", num);
   hexstr = add_separators(hextmp, thousep, 4);
 
   /* Find out the human-readable magnitude, e.g. 15.99 Mi */
@@ -1064,16 +1069,16 @@ terminal_util_timestamp_info (const char *str)
     return nullptr;
 
   /* Java uses Unix time in milliseconds. */
-  if (num >= 1000000000000 && num <= 1999999999999)
+  if (num >= 1000000000000 && num <= 9999999999999)
     num /= 1000;
 
   /* Fun: use inclusive interval so you can right-click on these numbers
-   * and check the human-readable time in gnome-terminal.
-   * (They're Sep 9 2001 and May 18 2033 by the way.) */
-  if (num < 1000000000 || num > 1999999999)
+   * right here and check the human-readable time in gnome-terminal.
+   * (They're Sep 9 2001 and Nov 20 2286 by the way.) */
+  if (num < 1000000000 || num > 9999999999)
     return nullptr;
 
-  gs_unref_date_time GDateTime* date = g_date_time_new_from_unix_utc (num);
+  gs_unref_date_time GDateTime* date = g_date_time_new_from_unix_local (num);
   if (date == nullptr)
     return nullptr;
 
@@ -1219,8 +1224,10 @@ ensure_cache_dir (void)
   cache_dir = get_cache_dir ();
   errno = 0;
   r = g_mkdir_with_parents (cache_dir, 0700);
-  if (r == -1 && errno != EEXIST)
-    g_printerr ("Failed to create cache dir: %m\n");
+  if (r == -1 && errno != EEXIST) {
+    auto const errsv = errno;
+    g_printerr ("Failed to create cache dir: %s\n", g_strerror(errsv));
+  }
   return r == 0;
 }
 
@@ -1568,366 +1575,112 @@ terminal_util_check_envv(char const* const* strv)
   return TRUE;
 }
 
-#define TERMINAL_SCHEMA_VERIFIER_ERROR (g_quark_from_static_string("TerminalSchemaVerifier"))
-
-typedef enum {
-  TERMINAL_SCHEMA_VERIFIER_SCHEMA_MISSING,
-  TERMINAL_SCHEMA_VERIFIER_SCHEMA_PATH,
-  TERMINAL_SCHEMA_VERIFIER_KEY_MISSING,
-  TERMINAL_SCHEMA_VERIFIER_KEY_TYPE,
-  TERMINAL_SCHEMA_VERIFIER_KEY_DEFAULT,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_UNKNOWN,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_MISMATCH,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_ENUM_VALUE,
-  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_INTERVAL,
-  TERMINAL_SCHEMA_VERIFIER_CHILD_MISSING,
-} TerminalSchemaVerifierError;
-
-static gboolean
-strv_contains(char const* const* strv,
-              char const* str)
+char**
+terminal_util_get_desktops(void)
 {
-  if (strv == nullptr)
-    return FALSE;
+  auto const desktop = g_getenv("XDG_CURRENT_DESKTOP");
+  if (!desktop)
+    return nullptr;
 
-  for (size_t i = 0; strv[i]; i++) {
-    if (g_str_equal (strv[i], str))
-      return TRUE;
-  }
-
-  return FALSE;
+  return g_strsplit(desktop, G_SEARCHPATH_SEPARATOR_S, -1);
 }
 
-static gboolean
-schema_key_range_compatible(GSettingsSchema* source_schema,
-                            GSettingsSchemaKey* source_key,
-                            char const* key,
-                            GSettingsSchemaKey* reference_key,
-                            GError** error)
+void
+terminal_util_menu_append_numbered (GMenu *menu,
+                                    const char *label,
+                                    int num,
+                                    const char *action_name,
+                                    GVariant *target /* consumed if floating */)
 {
-  gs_unref_variant GVariant* source_range =
-    g_settings_schema_key_get_range(source_key);
-  gs_unref_variant GVariant* reference_range =
-    g_settings_schema_key_get_range(reference_key);
+  gs_free_gstring GString *str;
+  gs_unref_object GMenuItem *item;
+  const char *p;
 
-  char const* source_type = nullptr;
-  gs_unref_variant GVariant* source_data = nullptr;
-  g_variant_get(source_range, "(&sv)", &source_type, &source_data);
+  /* Who'd use more that 4 underscores in a profile name... */
+  str = g_string_sized_new (strlen (label) + 4 + 1 + 8);
 
-  char const* reference_type = nullptr;
-  gs_unref_variant GVariant* reference_data = nullptr;
-  g_variant_get(reference_range, "(&sv)", &reference_type, &reference_data);
+  if (num < 10)
+    g_string_append_printf (str, "_%Id. ", num);
+  else if (num < 36)
+    g_string_append_printf (str, "_%c. ",  (char)('A' + num - 10));
 
-  if (!g_str_equal(source_type, reference_type)) {
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE,
-                "Schema \"%s\" key \"%s\" has range type \"%s\" but reference range type is \"%s\"",
-                g_settings_schema_get_id(source_schema),
-                key, source_type, reference_type);
-    return FALSE;
+  /* Append the label with underscores elided */
+  for (p = label; *p; p++) {
+    if (*p == '_')
+      g_string_append (str, "__");
+    else
+      g_string_append_c (str, *p);
   }
 
-  if (g_str_equal(reference_type, "type"))
-    ; /* no constraints; this is fine */
-  else if (g_str_equal(reference_type, "enum")) {
-    size_t source_values_len = 0;
-    gs_free char const** source_values = g_variant_get_strv(source_data, &source_values_len);
-
-    size_t reference_values_len = 0;
-    gs_free char const** reference_values = g_variant_get_strv(reference_data, &reference_values_len);
-
-    /* Check that every enum value in source is valid according to the reference */
-    for (size_t i = 0; i < source_values_len; ++i) {
-      if (!strv_contains(reference_values, source_values[i])) {
-        g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                    TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_ENUM_VALUE,
-                    "Schema \"%s\" key \"%s\" has enum value \"%s\" not in reference schema",
-                    g_settings_schema_get_id(source_schema),
-                    key, source_values[i]);
-        return FALSE;
-      }
-    }
-  } else if (g_str_equal(reference_type, "flags")) {
-    /* Our schemas don't use flags. If that changes, need to implement this! */
-    g_assert_not_reached();
-  } else if (g_str_equal(reference_type, "range")) {
-    if (!g_variant_is_of_type(source_data,
-                              g_variant_get_type(reference_data))) {
-      char const* source_type_str = g_variant_get_type_string(source_data);
-      char const* reference_type_str = g_variant_get_type_string(reference_data);
-      g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_MISMATCH,
-                  "Schema \"%s\" key \"%s\" has range type \"%s\" but reference range type is \"%s\"",
-                  g_settings_schema_get_id(source_schema),
-                  key, source_type_str, reference_type_str);
-      return FALSE;
-    }
-
-    gs_unref_variant GVariant* reference_min = nullptr;
-    gs_unref_variant GVariant* reference_max = nullptr;
-    g_variant_get(reference_data, "(**)", &reference_min, &reference_max);
-
-    gs_unref_variant GVariant* source_min = nullptr;
-    gs_unref_variant GVariant* source_max = nullptr;
-    g_variant_get(source_data, "(**)", &source_min, &source_max);
-
-    /* The source interval must be contained within the reference interval */
-    if (g_variant_compare(source_min, reference_min) < 0 ||
-        g_variant_compare(source_max, reference_max) > 0) {
-      g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                  TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_INTERVAL,
-                  "Schema \"%s\" key \"%s\" has range interval not contained in reference range interval",
-                  g_settings_schema_get_id(source_schema), key);
-        return FALSE;
-    }
-  } else {
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_KEY_RANGE_TYPE_UNKNOWN,
-                "Schema \"%s\" key \"%s\" has unknown range type \"%s\"",
-                g_settings_schema_get_id(source_schema),
-                key, reference_type);
-    return FALSE;
-  }
-
-  return TRUE;
+  item = g_menu_item_new (str->str, nullptr);
+  g_menu_item_set_action_and_target_value (item, action_name, target);
+  g_menu_item_set_attribute(item, "accel", "s", "");
+  g_menu_append_item (menu, item);
 }
 
-static gboolean
-schema_verify_key(GSettingsSchema* source_schema,
-                  char const* key,
-                  GSettingsSchema* reference_schema,
-                  GError** error)
+#define SETTINGS_ID "Terminal::Settings"
+#define KEY_ID "Terminal::Key"
+
+void
+terminal_util_set_settings_and_key_for_widget(GtkWidget* widget,
+                                              GSettings* settings,
+                                              char const* key)
 {
-  if (!g_settings_schema_has_key(source_schema, key)) {
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_KEY_MISSING,
-                "Schema \"%s\" has missing key \"%s\"",
-                g_settings_schema_get_id(source_schema), key);
-    return FALSE;
-  }
-
-  gs_unref_settings_schema_key GSettingsSchemaKey* source_key =
-    g_settings_schema_get_key(source_schema, key);
-  g_assert_nonnull(source_key);
-
-  gs_unref_settings_schema_key GSettingsSchemaKey* reference_key =
-    g_settings_schema_get_key(reference_schema, key);
-  g_assert_nonnull(reference_key);
-
-  GVariantType const* source_type = g_settings_schema_key_get_value_type(source_key);
-  GVariantType const* reference_type = g_settings_schema_key_get_value_type(reference_key);
-  if (!g_variant_type_equal(source_type, reference_type)) {
-    gs_free char* source_type_str = g_variant_type_dup_string(source_type);
-    gs_free char* reference_type_str = g_variant_type_dup_string(reference_type);
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_KEY_TYPE,
-                "Schema \"%s\" has type \"%s\" but reference type is \"%s\"",
-                g_settings_schema_get_id(source_schema),
-                source_type_str, reference_type_str);
-    return FALSE;
-  }
-
-  gs_unref_variant GVariant* source_default = g_settings_schema_key_get_default_value(source_key);
-  if (!g_settings_schema_key_range_check(reference_key, source_default)) {
-    gs_free char* source_value_str = g_variant_print(source_default, TRUE);
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_KEY_DEFAULT,
-                "Schema \"%s\" default value \"%s\" does not conform to reference schema",
-                g_settings_schema_get_id(source_schema), source_value_str);
-    return FALSE;
-  }
-
-  if (!schema_key_range_compatible(source_schema,
-                                   source_key,
-                                   key,
-                                   reference_key,
-                                   error))
-    return FALSE;
-
-  return TRUE;
+  g_return_if_fail(GTK_IS_WIDGET(widget));
+  g_return_if_fail(key);
+  g_return_if_fail(!terminal_util_get_settings_and_key_for_widget(widget, nullptr, nullptr));
+  g_object_set_data_full(G_OBJECT(widget), SETTINGS_ID,
+                         g_object_ref(settings),
+                         g_object_unref);
+  g_object_set_data(G_OBJECT(widget), KEY_ID,
+                    (void*)g_intern_string(key));
 }
 
-static gboolean
-schema_verify_child(GSettingsSchema* source_schema,
-                    char const* child_name,
-                    GSettingsSchema* reference_schema,
-                    GError** error)
+gboolean
+terminal_util_get_settings_and_key_for_widget(GtkWidget* widget,
+                                              GSettings** settings,
+                                              char const** key)
 {
-  /* Should verify the child's schema ID is as expected and exists in
-   * the source, but there appears to be no API to get the schema ID of
-   * the child.
-   *
-   * We work around this missing verification by never calling
-   * g_settings_get_child() and instead always constructing the child
-   * GSettings directly; and the existence and correctness of that
-   * schema is verified by the per-schema checks.
-   */
+  g_return_val_if_fail(GTK_IS_WIDGET(widget), false);
 
-  return TRUE;
+  auto const settings_data = reinterpret_cast<GSettings*>(g_object_get_data(G_OBJECT(widget), SETTINGS_ID));
+  auto const key_data = reinterpret_cast<char const*>(g_object_get_data(G_OBJECT(widget), KEY_ID));
+
+  if (settings)
+    *settings = settings_data;
+  if (key)
+    *key = key_data;
+
+  return settings_data && key_data;
 }
 
-static gboolean
-schema_verify(GSettingsSchema* source_schema,
-              GSettingsSchema* reference_schema,
-              GError** error)
+void
+terminal_util_g_settings_bind(GSettings* settings,
+                              char const* key,
+                              void* object,
+                              char const* property,
+                              GSettingsBindFlags flags)
 {
-  /* Verify path */
-  char const* source_path = g_settings_schema_get_path(source_schema);
-  char const* reference_path = g_settings_schema_get_path(reference_schema);
-  if (g_strcmp0(source_path, reference_path) != 0) {
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_SCHEMA_PATH,
-                "Schema \"%s\" has path \"%s\" but reference path is \"%s\"",
-                g_settings_schema_get_id(source_schema),
-                source_path ? source_path : "(null)",
-                reference_path ? reference_path : "(null)");
-    return FALSE;
-  }
+  g_return_if_fail(GTK_IS_WIDGET(object));
+  g_settings_bind(settings, key, object, property, flags);
 
-  /* Verify keys */
-  gs_strfreev char** keys = g_settings_schema_list_keys(reference_schema);
-  if (keys) {
-    for (int i = 0; keys[i]; ++i) {
-      if (!schema_verify_key(source_schema,
-                             keys[i],
-                             reference_schema,
-                             error))
-        return FALSE;
-    }
-  }
-
-  /* Verify child schemas */
-  gs_strfreev char** source_children = g_settings_schema_list_children(source_schema);
-  gs_strfreev char** reference_children = g_settings_schema_list_children(reference_schema);
-  if (reference_children) {
-    for (size_t i = 0; reference_children[i]; ++i) {
-      if (!strv_contains((char const* const*)source_children, reference_children[i])) {
-        g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                    TERMINAL_SCHEMA_VERIFIER_CHILD_MISSING,
-                    "Schema \"%s\" has missing child \"%s\"",
-                    g_settings_schema_get_id(source_schema),
-                    reference_children[i]);
-        return FALSE;
-      }
-
-      if (!schema_verify_child(source_schema,
-                               reference_children[i],
-                               reference_schema,
-                               error))
-          return FALSE;
-    }
-  }
-
-  return TRUE;
+  terminal_util_set_settings_and_key_for_widget(GTK_WIDGET(object), settings, key);
 }
 
-static gboolean
-schemas_source_verify_schema_by_name(GSettingsSchemaSource* source,
-                                     char const* schema_name,
-                                     GSettingsSchemaSource* reference_source,
-                                     GError** error)
+void
+terminal_util_g_settings_bind_with_mapping(GSettings* settings,
+                                           char const* key,
+                                           void* object,
+                                           char const* property,
+                                           GSettingsBindFlags flags,
+                                           GSettingsBindGetMapping get_mapping,
+                                           GSettingsBindSetMapping set_mapping,
+                                           void* user_data,
+                                           GDestroyNotify destroy)
 {
-  gs_unref_settings_schema GSettingsSchema* source_schema =
-    g_settings_schema_source_lookup(source, schema_name, TRUE /* recursive */);
+  g_return_if_fail(GTK_IS_WIDGET(object));
+  g_settings_bind_with_mapping(settings, key, object, property, flags,
+                               get_mapping, set_mapping, user_data, destroy);
 
-  if (!source_schema) {
-    g_set_error(error, TERMINAL_SCHEMA_VERIFIER_ERROR,
-                TERMINAL_SCHEMA_VERIFIER_SCHEMA_MISSING,
-                "Schema \"%s\" is missing", schema_name);
-    return FALSE;
-  }
-
-  gs_unref_settings_schema GSettingsSchema* reference_schema =
-    g_settings_schema_source_lookup(reference_source,
-                                    schema_name,
-                                    FALSE /* recursive */);
-  g_assert_nonnull(reference_schema);
-
-  return schema_verify(source_schema,
-                       reference_schema,
-                       error);
-}
-
-static gboolean
-schemas_source_verify_schemas(GSettingsSchemaSource* source,
-                              char const* const* schemas,
-                              GSettingsSchemaSource* reference_source,
-                              GError** error)
-{
-  if (!schemas)
-    return TRUE;
-
-  for (int i = 0; schemas[i]; ++i) {
-    if (!schemas_source_verify_schema_by_name(source,
-                                              schemas[i],
-                                              reference_source,
-                                              error))
-      return FALSE;
-  }
-
-  return TRUE;
-}
-
-static gboolean
-schemas_source_verify(GSettingsSchemaSource* source,
-                      GSettingsSchemaSource* reference_source,
-                      GError** error)
-{
-  gs_strfreev char** reloc_schemas = nullptr;
-  gs_strfreev char** nonreloc_schemas = nullptr;
-
-  g_settings_schema_source_list_schemas(reference_source,
-                                        FALSE /* recursive */,
-                                        &reloc_schemas,
-                                        &nonreloc_schemas);
-
-  if (!schemas_source_verify_schemas(source,
-                                     (char const* const*)reloc_schemas,
-                                     reference_source,
-                                     error))
-    return FALSE;
-
-  if (!schemas_source_verify_schemas(source,
-                                     (char const* const*)nonreloc_schemas,
-                                     reference_source,
-                                     error))
-    return FALSE;
-
-  return TRUE;
-}
-
-
-GSettingsSchemaSource*
-terminal_g_settings_schema_source_get_default(void)
-{
-  GSettingsSchemaSource* default_source = g_settings_schema_source_get_default();
-
-  gs_free_error GError* error = nullptr;
-  GSettingsSchemaSource* reference_source =
-    g_settings_schema_source_new_from_directory(TERM_PKGLIBDIR,
-                                                nullptr /* parent source */,
-                                                FALSE /* trusted */,
-                                                &error);
-  if (!reference_source)  {
-    /* Can only use the installed schemas, or abort here. */
-    g_printerr("Failed to load reference schemas: %s\n"
-               "Using unverified installed schemas.\n",
-               error->message);
-
-    return g_settings_schema_source_ref(default_source);
-  }
-
-  if (!schemas_source_verify(default_source, reference_source, &error)) {
-    g_printerr("Installed schemas failed verification: %s\n"
-               "Falling back to built-in reference schemas.\n",
-               error->message);
-
-    return reference_source; /* transfer */
-  }
-
-  /* Installed schemas verified; use them. */
-  g_settings_schema_source_unref(reference_source);
-  return g_settings_schema_source_ref(default_source);
+  terminal_util_set_settings_and_key_for_widget(GTK_WIDGET(object), settings, key);
 }
