@@ -448,6 +448,8 @@ vte_accessible_text_get_contents_at (GtkAccessibleText            *accessible,
         contents = &state->contents[state->contents_flip];
 
         if (contents->string == nullptr) {
+                *start = 0;
+                *end = 0;
                 return nullptr;
         }
 
@@ -471,7 +473,7 @@ vte_accessible_text_get_contents_at (GtkAccessibleText            *accessible,
                      line++) {
                         guint line_offset = *char_positions_index (&contents->linebreaks, line);
 
-                        if (line_offset > char_offset) {
+                        if (line_offset > offset) {
                                 line--;
                                 break;
                         }
@@ -493,20 +495,29 @@ vte_accessible_text_get_contents_at (GtkAccessibleText            *accessible,
         case GTK_ACCESSIBLE_TEXT_GRANULARITY_WORD: {
                 gunichar ch = vte_accessible_text_contents_get_char_at (contents, offset);
 
-                if (ch == 0 || !impl->is_word_char (ch))
+                if (ch == 0)
                         break;
-
+                if (!impl->is_word_char (ch)) {
+                        /* Find the end of the previous word, updating the offset to this positio n*/
+                        while (offset >= 0 &&
+                               (ch = vte_accessible_text_contents_get_char_at (contents, offset)) &&
+                               !impl->is_word_char (ch)) {
+                                offset--;
+                        }
+                }
                 *start = offset;
                 *end = offset;
 
-                while (*start > 0 &&
-                       (ch = vte_accessible_text_contents_get_char_at (contents, *start - 1)) &&
+                while (*start >= 0 &&
+                       (ch = vte_accessible_text_contents_get_char_at (contents, *start)) &&
                        impl->is_word_char (ch)) {
                         (*start)--;
                 }
+                /* Now, *start points one char before the real word start offset, so adjust it */
+                (*start)++;
 
                 while (*end < contents->n_chars &&
-                       (ch = vte_accessible_text_contents_get_char_at (contents, *end + 1)) &&
+                       (ch = vte_accessible_text_contents_get_char_at (contents, *end)) &&
                        impl->is_word_char (ch)) {
                         (*end)++;
                 }
@@ -520,6 +531,8 @@ vte_accessible_text_get_contents_at (GtkAccessibleText            *accessible,
                 break;
         }
 
+        *start = 0;
+        *end = 0;
         return nullptr;
 }
 
